"""some yams command line tools"""

import sys
from os.path import exists, join, dirname
from tempfile import NamedTemporaryFile
from subprocess import Popen
from traceback import extract_tb

from logilab.common.configuration import Configuration

from yams.__pkginfo__ import version
from yams import schema2dot
from yams.reader import SchemaLoader
from yams._exceptions import *

def _error(file=None, line=None, msg=''):
    if file is None:
        file = sys.argv[1]
    if line is None:
        line=''
    else:
        line = str(line)
    print >> sys.stderr, ':'.join(('E', file, line, msg))

def check_schema():
    config = Configuration(
        usage="yams-check [[[...] deps] deps] apps",
        doc="Check the schema of an application.",
        version=version)
    dirnames = config.load_command_line_configuration()
    if not dirnames:
        print >> sys.stderr, config.help()
        return return_code
    for dir in dirnames:
        assert exists(dir), dir
    try:
        SchemaLoader().load(dirnames)
        return 0
    except Exception, ex:
        tb_offset = getattr(ex, 'tb_offset', 0)
        filename, lineno, func, text = extract_tb(sys.exc_traceback)[-1-tb_offset]
        if hasattr(ex, "schema_files"):
            filename = ', '.join(ex.schema_files)
        _error(filename, lineno,"%s -> %s" % (ex.__class__.__name__, ex))
        return 2

def schema_image():
    options = [('output-file', {'type':'file', 'default': None,
                 'metavar': '<file>', 'short':'o', 'help':'output image file'}),
               ('viewer', {'type': 'string', 'default':"rsvg-view", 'short': "w", 'metavar':'<cmd>',
                 'help':'command use to view the generated file (empty for none)'}),
               ('lib-dir', {'type': 'string', 'short': "L", 'metavar':'<dir>',
                 'help':'directory to look for schema dependancies'}),
              ]

    config = Configuration(options=options,
        usage="yams-view [-L <lib_dir> | [[[...] deps] deps]] apps",
        version=version)
    dirnames = config.load_command_line_configuration()

    lib_dir = config['lib-dir']
    assert lib_dir is not None
    if lib_dir is not None:
        app_dir = dirnames[-1]
        from cubicweb.cwconfig import CubicWebConfiguration as cwcfg
        packages = cwcfg.expand_cubes(dirnames)
        packages = cwcfg.reorder_cubes(packages)
        packages = [pkg for pkg in packages if pkg != app_dir]
    elif False:
        glob = globals().copy()
        execfile(join(app_dir,"__pkginfo__.py"), glob)
        #dirnames = [ join(lib_dir,dep) for dep in glob['__use__']]+dirnames
        packages = [ dep for dep in glob['__use__']]


    for dir in dirnames:
        assert exists(dir), dir

    import cubicweb
    cubicweb_dir = dirname(cubicweb.__file__)

    schm_ldr = SchemaLoader()

    class MockConfig(object):
        def packages(self):
            return packages
        def packages_path(self):
            return packages
        def schemas_lib_dir(self):
            return join(cubicweb_dir,"schemas")
        #def apphome(self):
        #    return lib_dir
        apphome = dirnames[0]
        def appid(self):
            "bob"

    print MockConfig().packages()

    schema = schm_ldr.load(MockConfig())


    out, viewer = config['output-file'], config['viewer']
    if out is None:
        tmp_file = NamedTemporaryFile(suffix=".svg")
        out = tmp_file.name
    schema2dot.schema2dot(schema, out, #size=size,
        skiprels=("identity",), skipentities=("Person","AbstractPerson","Card","AbstractCompany","Company","Division"))
    if viewer:
        p = Popen((viewer, out))
        p.wait()

    return 0



# copyright 2004-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""unittests for schema2dot"""

import os
import os.path as osp

from logilab.common.testlib import TestCase, unittest_main

from yams.reader import SchemaLoader
from yams import schema2dot

DATADIR = osp.abspath(osp.join(osp.dirname(__file__), 'data'))

schema = SchemaLoader().load([DATADIR])

DOT_SOURCE = r'''digraph "toto" {
rankdir=BT
ratio=compress
charset="utf-8"
splines=true
overlap=false
sep=0.2
"Person" [fontname="Courier", label="{Person|nom\lprenom\lsexe\lpromo\ltitre\ladel\lass\lweb\ltel\lfax\ldatenaiss\ltest\lsalary\l}", shape="record", style="filled"];
"Salaried" [fontname="Courier", label="{Salaried|nom\lprenom\lsexe\lpromo\ltitre\ladel\lass\lweb\ltel\lfax\ldatenaiss\ltest\lsalary\l}", shape="record", style="filled"];
"Societe" [fontname="Courier", label="{Societe|nom\lweb\ltel\lfax\lrncs\lad1\lad2\lad3\lcp\lville\l}", shape="record", style="filled"];
"Salaried" -> "Person" [arrowhead="empty", color="grey", decorate="false", fontcolor="grey", label="Parent class", style="filled"];
"Person" -> "Societe" [arrowhead="normal", arrowtail="none", color="#aa0000", decorate="false", fontcolor="#aa0000", headlabel="0..n", label="travaille", style="filled", taillabel="0..n"];
"Salaried" -> "Societe" [arrowhead="normal", arrowtail="none", color="#00aa00", decorate="false", fontcolor="#00aa00", headlabel="0..n", label="travaille", style="filled", taillabel="0..n"];
}
'''

class DotTC(TestCase):

    def test_schema2dot(self):
        """tests dot conversion without attributes information"""
        wanted_entities = set(('Person', 'Salaried', 'Societe'))
        skipped_entities = set(ent for ent in schema.entities()
                               if ent.type not in wanted_entities)
        schema2dot.schema2dot(schema, 'toto.dot', skiptypes=skipped_entities)
        generated = open('toto.dot').read()
        os.remove('toto.dot')
        self.assertMultiLineEqual(DOT_SOURCE, generated)

if __name__ == '__main__':
    unittest_main()



# copyright 2004-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""Classes used to build a schema."""

__docformat__ = "restructuredtext en"

from warnings import warn
from copy import copy

from logilab.common import attrdict
from logilab.common.decorators import iclassmethod

from yams import BASE_TYPES, MARKER, BadSchemaDefinition, KNOWN_METAATTRIBUTES
from yams.constraints import (SizeConstraint, UniqueConstraint,
                              StaticVocabularyConstraint, FORMAT_CONSTRAINT)
from yams.schema import RelationDefinitionSchema

PACKAGE = '<builtin>' # will be modified by the yams'reader when schema is
                      # beeing read

__all__ = ('EntityType', 'RelationType', 'RelationDefinition',
           'SubjectRelation', 'ObjectRelation',
           'RichString', ) + tuple(BASE_TYPES)

# EntityType properties
ETYPE_PROPERTIES = ('description', '__permissions__', '__unique_together__')
# RelationType properties. Don't put description inside, handled specifically
RTYPE_PROPERTIES = ('symmetric', 'inlined', 'fulltext_container')
# RelationDefinition properties have to be computed dynamically since new ones
# may be added at runtime
def _RDEF_PROPERTIES():
    base = RelationDefinitionSchema.ALL_PROPERTIES()
    # infered is an internal property and should not be specified explicitly
    base.remove('infered')
    # replace permissions by __permissions__ as it's spelled that way in schema
    # definition files
    base.remove('permissions')
    base.add('__permissions__')
    return tuple(base)
# regroup all rtype/rdef properties as they may be defined one on each other in
# some cases
def _REL_PROPERTIES():
    return RTYPE_PROPERTIES + _RDEF_PROPERTIES()

# pre 0.37 backward compat
RDEF_PROPERTIES = () # stuff added here is also added to underlying dict, nevermind


CREATION_RANK = 0

def _add_constraint(kwargs, constraint):
    """Add constraint to param kwargs."""
    constraints = kwargs.setdefault('constraints', [])
    for i, existingconstraint in enumerate(constraints):
        if existingconstraint.__class__ is constraint.__class__:
            constraints[i] = constraint
            return
    constraints.append(constraint)

def _add_relation(relations, rdef, name=None, insertidx=None):
    """Add relation (param rdef) to list of relations (param relations)."""
    if name is not None:
        rdef.name = name
    if insertidx is None:
        insertidx = len(relations)
    relations.insert(insertidx, rdef)
    if getattr(rdef, 'metadata', {}):
        for meta_name, value in rdef.metadata.iteritems():
            assert meta_name in KNOWN_METAATTRIBUTES
            insertidx += 1 # insert meta after main
            meta_rel_name = '_'.join(((name or rdef.name), meta_name))
            _add_relation(relations, value, meta_rel_name, insertidx)

def _check_kwargs(kwargs, attributes):
    """Check that all keys of kwargs are actual attributes."""
    for key in kwargs:
        if not key in attributes:
            raise BadSchemaDefinition('no such property %r in %r'
                                      % (key, attributes))

def _copy_attributes(fromobj, toobj, attributes):
    for attr in attributes:
        value = getattr(fromobj, attr, MARKER)
        if value is MARKER:
            continue
        ovalue = getattr(toobj, attr, MARKER)
        if not ovalue is MARKER and value != ovalue:
            rname = getattr(toobj, 'name', None) or toobj.__name__
            raise BadSchemaDefinition(
                'conflicting values %r/%r for property %s of relation %r'
                % (ovalue, value, attr, rname))
        setattr(toobj, attr, value)

def register_base_types(schema):
    """add base (final) entity types to the given schema"""
    for etype in BASE_TYPES:
        edef = EntityType(name=etype)
        schema.add_entity_type(edef)


DEFAULT_RELPERMS = {'read': ('managers', 'users', 'guests',),
                    'delete': ('managers', 'users'),
                    'add': ('managers', 'users',)}

DEFAULT_ATTRPERMS = {'read': ('managers', 'users', 'guests',),
                     'add': ('managers', 'users'),
                     'update': ('managers', 'owners')}


# first class schema definition objects #######################################

class autopackage(type):
    def __new__(mcs, name, bases, classdict):
        classdict['package'] = PACKAGE
        return super(autopackage, mcs).__new__(mcs, name, bases, classdict)

class Definition(object):
    """Abstract class for entity / relation definition classes."""
    __metaclass__ = autopackage

    meta = MARKER
    description = MARKER
    __permissions__ = MARKER

    def __init__(self, name=None):
        self.name = (name or getattr(self, 'name', None)
                     or self.__class__.__name__)
        if self.__doc__:
            self.description = ' '.join(self.__doc__.split())

    def __repr__(self):
        return '<%s %r @%x>' % (self.__class__.__name__, self.name, id(self))

    @classmethod
    def expand_type_definitions(cls, defined):
        """Schema building step 1: register definition objects by adding them
        to the `defined` dictionnary.
        """
        raise NotImplementedError()

    @classmethod
    def expand_relation_definitions(cls, defined, schema):
        """Schema building step 2: register all relations definition,
        expanding wildcard if necessary.
        """
        raise NotImplementedError()

    @iclassmethod
    def get_permissions(cls, final=False):
        if cls.__permissions__ is MARKER:
            if final:
                return DEFAULT_ATTRPERMS
            return DEFAULT_RELPERMS
        return cls.__permissions__

    @classmethod
    def set_permissions(cls, perms):
        cls.__permissions__ = perms

    @classmethod
    def set_action_permissions(cls, action, actionperms):
        permissions = cls.get_permissions().copy()
        permissions[action] = actionperms
        cls.__permissions__ = permissions


# classes used to define relationships within entity type classes ##################

# has to be defined before the metadefinition metaclass which "isinstance" this
# class
class ObjectRelation(object):
    __permissions__ = MARKER
    cardinality = MARKER
    constraints = MARKER

    def __init__(self, etype, **kwargs):
        if self.__class__.__name__ == 'ObjectRelation':
            warn('[yams 0.29] ObjectRelation is deprecated, '
                 'use RelationDefinition subclass', DeprecationWarning,
                 stacklevel=2)
        global CREATION_RANK
        CREATION_RANK += 1
        self.creation_rank = CREATION_RANK
        self.package = PACKAGE
        self.name = '<undefined>'
        self.etype = etype
        if self.constraints:
            self.constraints = list(self.constraints)
        self.override = kwargs.pop('override', False)
        if kwargs.pop('meta', None):
            warn('[yams 0.37.0] meta is deprecated',
                 DeprecationWarning, stacklevel=3)
        try:
            _check_kwargs(kwargs, _REL_PROPERTIES())
        except BadSchemaDefinition, bad:
            # XXX (auc) bad field name + required attribute can lead there instead of schema.py ~ 920
            bsd_ex = BadSchemaDefinition(('%s in relation to entity %r (also is %r defined ? (check two '
                                          'lines above in the backtrace))') % (bad.args, etype, etype))
            bsd_ex.tb_offset = 2
            raise bsd_ex
        self.__dict__.update(kwargs)

    def __repr__(self):
        return '%(name)s %(etype)s' % self.__dict__


class SubjectRelation(ObjectRelation):
    uid = MARKER
    indexed = MARKER
    fulltextindexed = MARKER
    internationalizable = MARKER
    default = MARKER

    def __repr__(self):
        return '%(etype)s %(name)s' % self.__dict__


class AbstractTypedAttribute(SubjectRelation):
    """AbstractTypedAttribute is not directly instantiable

    subclasses must provide a <etype> attribute to be instantiable
    """
    def __init__(self, metadata=None, **kwargs):
        # Store metadata
        if metadata is None:
            metadata = {}
        self.metadata = metadata
        # transform "required" into "cardinality"
        required = kwargs.pop('required', False)
        if required:
            cardinality = '11'
        else:
            cardinality = '?1'
        kwargs['cardinality'] = cardinality
        # transform maxsize into SizeConstraint
        maxsize = kwargs.pop('maxsize', None)
        if maxsize is not None:
            _add_constraint(kwargs, SizeConstraint(max=maxsize))
        # transform vocabulary into StaticVocabularyConstraint
        vocabulary = kwargs.pop('vocabulary', None)
        if vocabulary is not None:
            self.set_vocabulary(vocabulary, kwargs)
        # transform unique into UniqueConstraint
        unique = kwargs.pop('unique', None)
        if unique:
            _add_constraint(kwargs, UniqueConstraint())
        # use the etype attribute provided by subclasses
        super(AbstractTypedAttribute, self).__init__(self.etype, **kwargs)
        # reassign creation rank
        #
        # Main attribute are marked as created before it's metadata.
        # order in meta data is preserved.
        if self.metadata:
            meta = sorted(metadata.values(), key= lambda x: x.creation_rank)
            if meta[0].creation_rank < self.creation_rank:
                m_iter = iter(meta)
                previous = self
                for next in meta:
                    if previous.creation_rank < next.creation_rank:
                        break
                    previous.creation_rank, next.creation_rank = next.creation_rank, previous.creation_rank
                    next = previous

    def set_vocabulary(self, vocabulary, kwargs=None):
        if kwargs is None:
            kwargs = self.__dict__
        #constraints = kwargs.setdefault('constraints', [])
        _add_constraint(kwargs, StaticVocabularyConstraint(vocabulary))
        if self.__class__.__name__ == 'String': # XXX
            maxsize = max(len(x) for x in vocabulary)
            _add_constraint(kwargs, SizeConstraint(max=maxsize))

    def __repr__(self):
        return '<%(name)s(%(etype)s)>' % self.__dict__


def make_type(etype):
    """create a python class for a Yams base type.

    Notice it is now possible to create a specific type with user-defined
    behaviour, e.g.:

        Geometry = make_type('Geometry') # (c.f. postgis)

    will allow the use of:

        Geometry(geom_type='POINT')

    in a Yams schema, provided in this example that `geom_type` is specified to
    the :func:`yams.register_base_type` function which should be called prior to
    make_type.
    """
    assert etype in BASE_TYPES
    return type(etype, (AbstractTypedAttribute,), {'etype' : etype})


# build a specific class for each base type
String = make_type('String')
Password = make_type('Password')
Bytes = make_type('Bytes')
Int = make_type('Int')
BigInt = make_type('BigInt')
Float = make_type('Float')
Boolean = make_type('Boolean')
Decimal = make_type('Decimal')
Time = make_type('Time')
Date = make_type('Date')
Datetime = make_type('Datetime')
TZTime = make_type('TZTime')
TZDatetime = make_type('TZDatetime')
Interval = make_type('Interval')


# provides a RichString factory for convenience
def RichString(default_format='text/plain', format_constraints=None, **kwargs):
    """RichString is a convenience attribute type for attribute containing text
    in a format that should be specified in another attribute.

    The following declaration::

      class Card(EntityType):
          content = RichString(fulltextindexed=True, default_format='text/rest')

    is equivalent to::

      class Card(EntityType):
          content_format = String(internationalizable=True,
                                  default='text/rest', constraints=[FORMAT_CONSTRAINT])
          content  = String(fulltextindexed=True)
    """
    format_args = {'default': default_format,
                   'maxsize': 50}
    if format_constraints is None:
        format_args['constraints'] = [FORMAT_CONSTRAINT]
    else:
        format_args['constraints'] = format_constraints
    meta = {'format':String(internationalizable=True, **format_args)}
    return String(metadata=meta, **kwargs)


# other schema definition classes ##############################################

class metadefinition(autopackage):
    """Metaclass that builds the __relations__ attribute of EntityType's
    subclasses.
    """
    stacklevel = 3
    def __new__(mcs, name, bases, classdict):
        ### Move (any) relation from the class dict to __relations__ attribute
        rels = classdict.setdefault('__relations__', [])
        relations = dict((rdef.name, rdef) for rdef in rels)
        for rname, rdef in classdict.items():
            if isinstance(rdef, ObjectRelation):
                # relation's name **must** be removed from class namespace
                # to avoid conflicts with instance's potential attributes
                del classdict[rname]
                relations[rname] = rdef
        ### handle logical inheritance
        if '__specializes_schema__' in classdict:
            specialized = bases[0]
            classdict['__specializes__'] = specialized.__name__
            if '__specialized_by__' not in specialized.__dict__:
                specialized.__specialized_by__ = []
            specialized.__specialized_by__.append(name)
        ### Initialize processed class
        defclass = super(metadefinition, mcs).__new__(mcs, name, bases, classdict)
        for rname, rdef in relations.items():
            _add_relation(defclass.__relations__, rdef, rname)
        ### take base classes'relations into account
        for base in bases:
            for rdef in getattr(base, '__relations__', ()):
                if not rdef.name in relations or not relations[rdef.name].override:
                    if isinstance(rdef, RelationDefinition):
                        rdef = copy(rdef)
                        if rdef.subject == base.__name__:
                            rdef.subject = name
                        if rdef.object == base.__name__:
                            rdef.object = name
                    rels.append(rdef)
                else:
                    relations[rdef.name].creation_rank = rdef.creation_rank
        ### sort relations by creation rank
        defclass.__relations__ = sorted(rels, key=lambda r: r.creation_rank)
        return defclass


class EntityType(Definition):
    #::FIXME reader magic forbids to define a docstring...
    #: an entity has attributes and can be linked to other entities by
    #: relations. Both entity attributes and relationships are defined by
    #: class attributes.
    #:
    #: kwargs keys must have values in ETYPE_PROPERTIES
    #:
    #: Example:
    #:
    #: >>> class Project(EntityType):
    #: ...     name = String()
    #: >>>
    #:
    #: After instanciation, EntityType can we altered with dedicated class methods:
    #:
    #: .. currentmodule:: yams.buildobjs
    #:
    #:  .. automethod:: EntityType.extend
    #:  .. automethod:: EntityType.add_relation
    #:  .. automethod:: EntityType.insert_relation_after
    #:  .. automethod:: EntityType.remove_relation
    #:  .. automethod:: EntityType.get_relation
    #:  .. automethod:: EntityType.get_relations

    __metaclass__ = metadefinition
    __permissions__ = {
        'read': ('managers', 'users', 'guests',),
        'update': ('managers', 'owners',),
        'delete': ('managers', 'owners'),
        'add': ('managers', 'users',)
        }

    def __init__(self, name=None, **kwargs):
        super(EntityType, self).__init__(name)
        _check_kwargs(kwargs, ETYPE_PROPERTIES)
        self.__dict__.update(kwargs)
        self.specialized_type = self.__class__.__dict__.get('__specializes__')

    def __str__(self):
        return 'entity type %r' % self.name

    @property
    def specialized_by(self):
        return self.__class__.__dict__.get('__specialized_by__', [])

    @classmethod
    def expand_type_definitions(cls, defined):
        """Schema building step 1: register definition objects by adding
        them to the `defined` dictionnary.
        """
        name = getattr(cls, 'name', cls.__name__)
        assert cls is not defined.get(name), 'duplicate registration: %s' % name
        assert name not in defined, \
            "type '%s' was already defined here %s, new definition here %s" % \
            (name, defined[name].__module__, cls)
        cls._defined = defined # XXX may be used later (eg .add_relation())
        defined[name] = cls
        for relation in cls.__relations__:
            cls._ensure_relation_type(relation)

    @classmethod
    def _ensure_relation_type(cls, relation):
        """Check the type the relation

        return False if the class is not yet finalized
        (XXX raise excep instead ?)"""
        rtype = RelationType(relation.name)
        _copy_attributes(relation, rtype, RTYPE_PROPERTIES)
        #assert hasattr(cls, '_defined'), "Type definition for %s not yet expanded. you can't register new type through it" % cls
        if hasattr(cls, '_defined'):
            defined = cls._defined
            if relation.name in defined:
                _copy_attributes(rtype, defined[relation.name], RTYPE_PROPERTIES)
            else:
                defined[relation.name] = rtype
            return True
        else:
            return False

    @classmethod
    def expand_relation_definitions(cls, defined, schema):
        """schema building step 2:

        register all relations definition, expanding wildcards if necessary
        """
        order = 1
        name = getattr(cls, 'name', cls.__name__)
        rdefprops = _RDEF_PROPERTIES()
        for relation in cls.__relations__:
            if isinstance(relation, SubjectRelation):
                rdef = RelationDefinition(subject=name, name=relation.name,
                                          object=relation.etype, order=order,
                                          package=relation.package)
                _copy_attributes(relation, rdef, rdefprops)
            elif isinstance(relation, ObjectRelation):
                rdef = RelationDefinition(subject=relation.etype,
                                          name=relation.name,
                                          object=name, order=order,
                                          package=relation.package)
                _copy_attributes(relation, rdef, rdefprops)
            elif isinstance(relation, RelationDefinition):
                rdef = relation
            else:
                raise BadSchemaDefinition('dunno how to handle %s' % relation)
            order += 1
            rdef._add_relations(defined, schema)

    # methods that can be used to extend an existant schema definition ########

    @classmethod
    def extend(cls, othermetadefcls):
        """add all relations of ``othermetadefcls`` to the current class"""
        for rdef in othermetadefcls.__relations__:
            cls.add_relation(rdef)

    @classmethod
    def add_relation(cls, rdef, name=None):
        """Add ``rdef`` relation to the class"""
        if name:
            rdef.name = name
        if cls._ensure_relation_type(rdef):
            _add_relation(cls.__relations__, rdef, name)
            if getattr(rdef, 'metadata', {}) and not rdef in cls._defined:
                for meta_name in rdef.metadata:
                    meta_rel_name = '_'.join(((name or rdef.name), name_name))
                    rdef = cls.get_relations(format_attr_name).next()
                    cls._ensure_relation_type(rdef)
        else:
           _add_relation(cls.__relations__, rdef, name=name)

    @classmethod
    def insert_relation_after(cls, afterrelname, name, rdef):
        """Add ``rdef`` relation to the class right after another"""
        # FIXME change order of arguments to rdef, name, afterrelname ?
        rdef.name = name
        cls._ensure_relation_type(rdef)
        for i, rel in enumerate(cls.__relations__):
            if rel.name == afterrelname:
                break
        else:
            raise BadSchemaDefinition("can't find %s relation on %s" % (
                    afterrelname, cls))
        _add_relation(cls.__relations__, rdef, name, i+1)

    @classmethod
    def remove_relation(cls, name):
        """Remove relation from the class"""
        for rdef in cls.get_relations(name):
            cls.__relations__.remove(rdef)

    @classmethod
    def get_relations(cls, name):
        """Iterate over relations definitions that match the ``name`` parameters

        It may iterate multiple definitions when the class is both object and
        sujet of a relation:
        """
        for rdef in cls.__relations__[:]:
            if rdef.name == name:
                yield rdef

    @classmethod
    def get_relation(cls, name):
        """Return relation definitions by name. Fails if there is multiple one.
        """
        relations = tuple(cls.get_relations(name))
        assert len(relations) == 1, "can't use get_relation for relation with multiple definitions"
        return relations[0]


class RelationType(Definition):
    symmetric = MARKER
    inlined = MARKER
    fulltext_container = MARKER

    def __init__(self, name=None, **kwargs):
        """kwargs must have values in RTYPE_PROPERTIES"""
        super(RelationType, self).__init__(name)
        if kwargs.pop('meta', None):
            warn('[yams 0.37] meta is deprecated', DeprecationWarning, stacklevel=2)
        _check_kwargs(kwargs, RTYPE_PROPERTIES + ('description', '__permissions__'))
        self.__dict__.update(kwargs)

    def __str__(self):
        return 'relation type %r' % self.name

    @classmethod
    def expand_type_definitions(cls, defined):
        """schema building step 1:

        register definition objects by adding them to the `defined` dictionnary
        """
        name = getattr(cls, 'name', cls.__name__)
        if cls.__doc__ and not cls.description:
            cls.description = ' '.join(cls.__doc__.split())
        if name in defined:
            if defined[name].__class__ is not RelationType:
                raise BadSchemaDefinition('duplicated relation type for %s'
                                          % name)
            # relation type created from a relation definition, override it
            allprops = _REL_PROPERTIES() + ('subject', 'object')
            _copy_attributes(defined[name], cls, allprops)
        defined[name] = cls

    @classmethod
    def expand_relation_definitions(cls, defined, schema):
        """schema building step 2:

        register all relations definition, expanding wildcard if necessary
        """
        name = getattr(cls, 'name', cls.__name__)
        if getattr(cls, 'subject', None) and getattr(cls, 'object', None):
            rdef = RelationDefinition(subject=cls.subject, name=name,
                                      object=cls.object)
            _copy_attributes(cls, rdef, _RDEF_PROPERTIES())
            rdef._add_relations(defined, schema)


class RelationDefinition(Definition):
    # FIXME reader magic forbids to define a docstring...
    #"""a relation is defined by a name, the entity types that can be
    #subject or object the relation, the cardinality, the constraints
    #and the symmetric property.
    #"""

    subject = MARKER
    object = MARKER
    cardinality = MARKER
    constraints = MARKER
    symmetric = MARKER
    inlined = MARKER

    def __init__(self, subject=None, name=None, object=None, package=None,
                 **kwargs):
        """kwargs keys must have values in _RDEF_PROPERTIES()"""
        if subject:
            self.subject = subject
        else:
            self.subject = self.__class__.subject
        if object:
            self.object = object
        else:
            self.object = self.__class__.object
        super(RelationDefinition, self).__init__(name)
        global CREATION_RANK
        CREATION_RANK += 1
        self.creation_rank = CREATION_RANK
        if package is not None:
            self.package = package
        elif self.package == '<builtin>':
            self.package = PACKAGE
        if kwargs.pop('meta', None):
            warn('[yams 0.37] meta is deprecated', DeprecationWarning)
        rdefprops = _RDEF_PROPERTIES()
        _check_kwargs(kwargs, rdefprops)
        _copy_attributes(attrdict(**kwargs), self, rdefprops)
        if self.constraints:
            self.constraints = list(self.constraints)

    def __str__(self):
        return 'relation definition (%(subject)s %(name)s %(object)s)' % self.__dict__

    @classmethod
    def expand_type_definitions(cls, defined):
        """schema building step 1:

        register definition objects by adding them to the `defined` dictionnary
        """
        name = getattr(cls, 'name', cls.__name__)
        rtype = RelationType(name)
        _copy_attributes(cls, rtype, RTYPE_PROPERTIES)
        if name in defined:
            _copy_attributes(rtype, defined[name], RTYPE_PROPERTIES)
        else:
            defined[name] = rtype

        # subject and object in defined's keys are only strings not tuples
        if isinstance(cls.subject, tuple):
            subjects = cls.subject
        else:
            subjects = (cls.subject, )
        if isinstance(cls.object, tuple):
            objects = cls.object
        else:
            objects = (cls.object, )
        for sub in subjects:
            for obj in objects:
                key = (sub, name, obj)
                if key in defined:
                    raise BadSchemaDefinition(
                        'duplicated relation definition (%s) %s (%s.%s)'
                        % (defined[key], key, cls.__module__, cls.__name__))
                defined[key] = cls

        # XXX keep this for bw compat
        defined[(cls.subject, name, cls.object)] = cls

    @classmethod
    def expand_relation_definitions(cls, defined, schema):
        """schema building step 2:

        register all relations definition, expanding wildcard if necessary
        """
        assert cls.subject and cls.object, '%s; check the schema (%s, %s)' % (cls, cls.subject, cls.object)
        cls()._add_relations(defined, schema)

    def _add_relations(self, defined, schema):
        name = getattr(self, 'name', self.__class__.__name__)
        rtype = defined[name]
        rdefprops = _RDEF_PROPERTIES()
        # copy relation definition attributes set on the relation type, beside
        # description
        _copy_attributes(rtype, self, set(rdefprops) - set(('description',)))
        # process default cardinality and constraints if not set yet
        cardinality = self.cardinality
        if cardinality is MARKER:
            if self.object in BASE_TYPES:
                self.cardinality = '?1'
            else:
                self.cardinality = '**'
        else:
            assert len(cardinality) == 2
            assert cardinality[0] in '1?+*'
            assert cardinality[1] in '1?+*'
        if not self.constraints:
            self.constraints = ()
        rschema = schema.rschema(name)
        if self.__permissions__ is MARKER:
            final = iter(_actual_types(schema, self.object)).next() in BASE_TYPES
            permissions = rtype.get_permissions(final)
        else:
            permissions = self.__permissions__
        for subj in _actual_types(schema, self.subject):
            for obj in _actual_types(schema, self.object):
                rdef = RelationDefinition(subj, name, obj,
                                          __permissions__=permissions,
                                          package=self.package)
                _copy_attributes(self, rdef, rdefprops)
                schema.add_relation_def(rdef)

def _actual_types(schema, etype):
    if etype == '*':
        return _pow_etypes(schema)
    if isinstance(etype, (list, tuple)):
        return etype
    if not isinstance(etype, basestring):
        raise RuntimeError('Entity types must not be instances but strings '
                           'or list/tuples thereof. Ex. (bad, good) : '
                           'SubjectRelation(Foo), SubjectRelation("Foo"). '
                           'Hence, %r is not acceptable.' % etype)
    return (etype,)

def _pow_etypes(schema):
    for eschema in schema.entities():
        if eschema.final:
            continue
        yield eschema.type



# copyright 2004-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""Object model and utilities to define generic Entities/Relations schemas.
"""
__docformat__ = "restructuredtext en"

import warnings
from datetime import datetime, date, time

# XXX set _ builtin to unicode by default, should be overriden if necessary
import __builtin__
__builtin__._ = unicode

from logilab.common.date import strptime, strptime_time
from logilab.common import nullobject

from yams.__pkginfo__ import version as __version__
from yams._exceptions import *

MARKER = nullobject()

BASE_TYPES = set(('String', 'Password', 'Bytes',
                  'Int', 'BigInt', 'Float', 'Boolean', 'Decimal',
                  'Date', 'Time', 'Datetime', 'TZTime', 'TZDatetime', 'Interval'
                  ))

# base groups used in permissions
BASE_GROUPS = set((_('managers'), _('users'), _('guests'), _('owners')))

# This provides a way to specify callable objects as default values
# First level is the final type, second is the keyword to callable map
KEYWORD_MAP = {
    'Datetime':{'NOW' : datetime.now,
                'TODAY': datetime.today},
    'TZDatetime': {'NOW' : datetime.utcnow,
                   'TODAY': datetime.today},
    'Date': {'TODAY': date.today}
}


# bw compat for literal date/time values stored as strings in schemas
DATE_FACTORY_MAP = {
    'Datetime' : lambda x: ':' in x and strptime(x, '%Y/%m/%d %H:%M') or strptime(x, '%Y/%m/%d'),
    'Date' : lambda x : strptime(x, '%Y/%m/%d'),
    'Time' : strptime_time
    }


def convert_default_value(rdef, default):
    # rdef can be either a yams.schema.RelationDefinitionSchema or a yams.buildobjs.RelationDefinition
    rtype = getattr(rdef, 'name', None) or rdef.rtype.type
    if isinstance(default, basestring) and rdef.object != 'String':
        # real Strings can be anything, including things that look like keywords
        # for other base types
        if rdef.object in KEYWORD_MAP:
            try:
                return KEYWORD_MAP[rdef.object][default.upper()]()
            except KeyError:
                # the default was likely not a special constant like TODAY but some literal
                pass
        # bw compat for old schemas
        if rdef.object in DATE_FACTORY_MAP:
            warnings.warn('using strings as default values for attribute %s of type %s '
                          'is deprecated; you should use the plain python objects instead'
                          % (rtype, rdef.object),
                          DeprecationWarning)
            try:
                return DATE_FACTORY_MAP[rdef.object](default)
            except ValueError, verr:
                raise ValueError('creating a default value for attribute %s of type %s '
                                 'from the string %r is not supported (cause %s)'
                                 % (rtype, rdef.object, default, verr))
    if rdef.object == 'String':
        default = unicode(default)
    return default # general case: untouched default



KNOWN_METAATTRIBUTES = set(('format', 'encoding', 'name'))

def register_base_type(name, parameters=(), check_function=None):
    """register a yams base (final) type. You'll have to call
    base_type_class to generate the class.
    """
    from yams.schema import RelationDefinitionSchema
    from yams.constraints import BASE_CHECKERS, yes
    # Add the datatype to yams base types
    assert name not in BASE_TYPES, '%s alreadt in BASE_TYPES %s' % (name, BASE_TYPES)
    BASE_TYPES.add(name)
    # Add the new datatype to the authorized types of RelationDefinitionSchema
    if not isinstance(parameters, dict):
        # turn tuple/list into dict with None values
        parameters = dict((p, None) for p in parameters)
    RelationDefinitionSchema.BASE_TYPE_PROPERTIES[name] = parameters
    # Add a yams checker or yes is not specified
    BASE_CHECKERS[name] = check_function or yes

def unregister_base_type(name):
    """Unregister a yams base (final) type"""
    from yams.schema import RelationDefinitionSchema
    from yams.constraints import BASE_CHECKERS
    assert name in BASE_TYPES, '%s not in BASE_TYPES %s' % (name, BASE_TYPES)
    BASE_TYPES.remove(name)
    RelationDefinitionSchema.BASE_TYPE_PROPERTIES.pop(name)
    BASE_CHECKERS.pop(name)



# copyright 2004-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""ER schema loader.

Use either a sql derivated language for entities and relation definitions
files or a direct python definition file.
"""

__docformat__ = "restructuredtext en"

import sys
import types
from os import listdir
from os.path import exists, join, splitext, basename, abspath
from warnings import warn

from logilab.common import tempattr
from logilab.common.modutils import modpath_from_file, cleanup_sys_modules

from yams import UnknownType, BadSchemaDefinition, BASE_TYPES
from yams import constraints, schema as schemamod
from yams import buildobjs


CONSTRAINTS = {}
# add constraint classes to the context
for objname in dir(constraints):
    if objname[0] == '_':
        continue
    obj = getattr(constraints, objname)
    try:
        if issubclass(obj, constraints.BaseConstraint) and (
            not obj is constraints.BaseConstraint):
            CONSTRAINTS[objname] = obj
    except TypeError:
        continue


def obsolete(cls):
    def wrapped(*args, **kwargs):
        reason = '%s should be explictly imported from %s' % (
            cls.__name__, cls.__module__)
        warn(reason, DeprecationWarning, stacklevel=2)
        return cls(*args, **kwargs)
    return wrapped

def fill_schema(schema, erdefs, register_base_types=True,
                remove_unused_rtypes=False, post_build_callbacks=[]):
    if register_base_types:
        buildobjs.register_base_types(schema)
    # relation definitions may appear multiple times
    erdefs_vals = set(erdefs.itervalues())
    # register relation types and non final entity types
    for definition in erdefs_vals:
        if isinstance(definition, type):
            definition = definition()
        if isinstance(definition, buildobjs.RelationType):
            schema.add_relation_type(definition)
        elif isinstance(definition, buildobjs.EntityType):
            schema.add_entity_type(definition)
    # register relation definitions
    for definition in erdefs_vals:
        if isinstance(definition, type):
            definition = definition()
        definition.expand_relation_definitions(erdefs, schema)
    # call 'post_build_callback' functions found in schema modules
    for cb in post_build_callbacks:
        cb(schema)
    # optionaly remove relation types without definitions
    if remove_unused_rtypes:
        for rschema in schema.relations():
            if not rschema.rdefs:
                schema.del_relation_type(rschema)
    # set permissions on entities and relations
    for erschema in schema.entities() + schema.relations():
        erschema.check_permission_definitions()
    schema.infer_specialization_rules()
    for eschema in schema.entities():
        eschema.check_unique_together()
    return schema


class SchemaLoader(object):
    """the schema loader is responsible to build a schema object from a
    set of files
    """
    schemacls = schemamod.Schema
    extrapath = None
    context = dict([(attr, getattr(buildobjs, attr))
                    for attr in buildobjs.__all__])
    context.update(CONSTRAINTS)

    def load(self, directories, name=None,
             register_base_types=True, construction_mode='strict',
             remove_unused_rtypes=True):
        """return a schema from the schema definition read from <directory>
        """
        self.defined = {}
        self.loaded_files = []
        self.post_build_callbacks = []
        sys.modules[__name__].context = self
        # ensure we don't have an iterator
        directories = tuple(directories)
        try:
            self._load_definition_files(directories)
            schema = self.schemacls(name or 'NoName', construction_mode=construction_mode)
            try:
                fill_schema(schema, self.defined, register_base_types,
                            remove_unused_rtypes=remove_unused_rtypes,
                            post_build_callbacks=self.post_build_callbacks)
            except Exception, ex:
                if not hasattr(ex, 'schema_files'):
                    ex.schema_files = self.loaded_files
                raise ex, None, sys.exc_info()[-1]
        finally:
            # cleanup sys.modules from schema modules
            # ensure we're only cleaning schema [sub]modules
            directories = [(not directory.endswith(self.main_schema_directory)
                            and join(directory, self.main_schema_directory)
                            or directory)
                           for directory in directories]
            cleanup_sys_modules(directories)
        schema.loaded_files = self.loaded_files
        return schema

    def _load_definition_files(self, directories):
        for directory in directories:
            package = basename(directory)
            for filepath in self.get_schema_files(directory):
                with tempattr(buildobjs, 'PACKAGE', package):
                    self.handle_file(filepath)

    # has to be overridable sometimes (usually for test purpose)
    main_schema_directory = 'schema'
    def get_schema_files(self, directory):
        """return an ordered list of files defining a schema

        look for a schema.py file and or a schema sub-directory in the given
        directory
        """
        result = []
        if exists(join(directory, self.main_schema_directory + '.py')):
            result = [join(directory, self.main_schema_directory + '.py')]
        if exists(join(directory, self.main_schema_directory)):
            directory = join(directory, self.main_schema_directory)
            for filename in listdir(directory):
                if filename[0] == '_':
                    if filename == '__init__.py':
                        result.insert(0, join(directory, filename))
                    continue
                ext = splitext(filename)[1]
                if ext == '.py':
                    result.append(join(directory, filename))
                else:
                    self.unhandled_file(join(directory, filename))
        return result

    def handle_file(self, filepath):
        """handle a partial schema definition file according to its extension
        """
        assert filepath.endswith('.py'), 'not a python file'
        if filepath not in self.loaded_files:
            modname, module = self.exec_file(filepath)
            objects_to_add = set()
            for name, obj in vars(module).items():
                if (isinstance(obj, type)
                    and issubclass(obj, buildobjs.Definition)
                    and obj.__module__ == modname
                    and not name.startswith('_')):
                    objects_to_add.add(obj)
            for obj in objects_to_add:
                self.add_definition(obj, filepath)
            self.loaded_files.append(filepath)

    def unhandled_file(self, filepath):
        """called when a file without handler associated has been found,
        does nothing by default.
        """
        pass

    def add_definition(self, defobject, filepath=None):
        """file handler callback to add a definition object

        wildcard capability force to load schema in two steps : first register
        all definition objects (here), then create actual schema objects (done in
        `_build_schema`)
        """
        if not issubclass(defobject, buildobjs.Definition):
            raise BadSchemaDefinition(filepath, 'invalid definition object')
        defobject.expand_type_definitions(self.defined)

    def import_erschema(self, ertype, schemamod=None, instantiate=True):
        warn('import_erschema is deprecated, use explicit import once schema '
             'is turned into a proper python module (eg not expecting '
             'predefined context in globals)', DeprecationWarning, stacklevel=3)
        try:
            erdef = self.defined[ertype]
            name = getattr(erdef, 'name', erdef.__name__)
            if name == ertype:
                assert instantiate, 'can\'t get class of an already registered type'
                return erdef
        except KeyError:
            pass
        assert False, 'ooups'

    def exec_file(self, filepath):
        try:
            modname = '.'.join(modpath_from_file(filepath, self.extrapath))
            doimport = True
        except ImportError:
            warn('module for %s can\'t be found, add necessary __init__.py '
                 'files to make it importable' % filepath, DeprecationWarning)
            modname = splitext(basename(filepath))[0]
            doimport = False
        # XXX can't rely on __import__ until bw compat (eg implicit import) needed
        #if doimport:
        #    module = __import__(modname, fglobals)
        #    for part in modname.split('.')[1:]:
        #        module = getattr(module, part)
        #else:
        if modname in sys.modules:
            module = sys.modules[modname]
            # NOTE: don't test raw equality to avoid .pyc / .py comparisons
            assert abspath(module.__file__).startswith(abspath(filepath)), (
                modname, filepath, module.__file__)
        else:
            # XXX until bw compat is gone, put context into builtins to allow proper
            # control of deprecation warning
            import __builtin__
            fglobals = {} # self.context.copy()
            # wrap callable that should be imported
            for key, val in self.context.items():
                if key in BASE_TYPES or key == 'RichString' or key in CONSTRAINTS or \
                       key in ('SubjectRelation', 'ObjectRelation'):
                    val = obsolete(val)
                setattr(__builtin__, key, val)
            __builtin__.import_erschema = self.import_erschema
            fglobals['__file__'] = filepath
            fglobals['__name__'] = modname
            package = '.'.join(modname.split('.')[:-1])
            if package and not package in sys.modules:
                __import__(package)
            execfile(filepath, fglobals)
            # check for use of classes that should be imported, without
            # importing them
            for name, obj in fglobals.items():
                if isinstance(obj, type) and \
                       issubclass(obj, buildobjs.Definition) and \
                       obj.__module__ == modname:
                    for parent in obj.__bases__:
                        pname = parent.__name__
                        if pname in fglobals or not pname in self.context:
                            # imported
                            continue
                        warn('%s: please explicitly import %s (%s)'
                             % (filepath, pname, name), DeprecationWarning)
            #for key in self.context:
            #    fglobals.pop(key, None)
            fglobals['__file__'] = filepath
            module = types.ModuleType(str(modname))
            module.__dict__.update(fglobals)
            sys.modules[modname] = module
            if package:
                setattr(sys.modules[package], modname.split('.')[-1], module)
        if hasattr(module, 'post_build_callback'):
            self.post_build_callbacks.append(module.post_build_callback)
        return (modname, module)

# XXX backward compatibility to prevent changing cw.schema and cw.test.unittest_schema (3.12.+)
PyFileReader = SchemaLoader
PyFileReader.__init__ = lambda *x: None

def build_schema_from_namespace(items):
    erdefs = {}
    for name, obj in items:
        if (isinstance(obj, type) and issubclass(obj, buildobjs.Definition)
            and obj not in (buildobjs.Definition, buildobjs.RelationDefinition, buildobjs.EntityType)):
            obj.expand_type_definitions(erdefs)
    schema = schemamod.Schema('noname')
    fill_schema(schema, erdefs)
    return schema

class _Context(object):
    def __init__(self):
        self.defined = {}

context = _Context()



# copyright 2004-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""Classes to define generic Entities/Relations schemas."""

__docformat__ = "restructuredtext en"
_ = unicode

import warnings
from copy import deepcopy
from decimal import Decimal
from itertools import chain

from logilab.common import attrdict
from logilab.common.decorators import cached, clear_cache
from logilab.common.interface import implements

import yams
from yams import (BASE_TYPES, MARKER, ValidationError, BadSchemaDefinition,
                  KNOWN_METAATTRIBUTES, convert_default_value)
from yams.interfaces import (ISchema, IRelationSchema, IEntitySchema,
                             IVocabularyConstraint)
from yams.constraints import BASE_CHECKERS, BASE_CONVERTERS, UniqueConstraint

def role_name(rtype, role):
    """function to use for qualifying attribute / relation in ValidationError
    errors'dictionnary
    """
    return '%s-%s' % (rtype, role)

def check_permission_definitions(schema):
    """check permissions are correctly defined"""
    # already initialized, check everything is fine
    for action, groups in schema.permissions.items():
        assert action in schema.ACTIONS, \
            'unknown action %s for %s' % (action, schema)
        assert isinstance(groups, tuple), \
               ('permission for action %s of %s isn\'t a tuple as '
                'expected' % (action, schema))
    if schema.final:
        schema.advertise_new_add_permission()
    for action in schema.ACTIONS:
        assert action in schema.permissions, \
               'missing expected permissions for action %s for %s' % (action, schema)

def rehash(dictionary):
    """this function manually builds a copy of `dictionary` but forces
    hash values to be recomputed. Note that dict(d) or d.copy() don't
    do that.

    It is used to :
      - circumvent Pyro / (un)pickle problems (hash mode is changed
        during reconstruction)
      - force to recompute keys' hash values. This is needed when a
        schema's type is changed because the schema's hash method is based
        on the type attribute. This problem is illusrated by the pseudo-code
        below :

        >>> topic = EntitySchema(type='Topic')
        >>> d = {topic : 'foo'}
        >>> d[topic]
        'foo'
        >>> d['Topic']
        'foo'
        >>> topic.type = 'folder'
        >>> topic in d
        False
        >>> 'Folder' in d
        False
        >>> 'Folder' in d.keys() # but it can be found "manually"
        True
        >>> d = rehash(d) # explicit rehash()
        >>> 'Folder' in d
        True
    """
    return dict(item for item in dictionary.items())

def _format_properties(props):
    res = [('%s=%s' % item) for item in props.items() if item[1]]
    return ','.join(res)

class ERSchema(object):
    """Base class shared by entity and relation schema."""


    def __init__(self, schema=None, erdef=None):
        """
        Construct an ERSchema instance.

        :Parameters:
         - `schema`: (??)
         - `erdef`: (??)
        """
        if erdef is None:
            return
        self.schema = schema
        self.type = erdef.name
        self.description = erdef.description or ''
        self.package = erdef.package

    def __cmp__(self, other):
        return cmp(self.type, getattr(other, 'type', other))

    def __hash__(self):
        try:
            return hash(self.type)
        except AttributeError:
            pass
        return hash(id(self))

    def __deepcopy__(self, memo):
        clone = self.__class__()
        memo[id(self)] = clone
        clone.type = deepcopy(self.type, memo)
        clone.schema = deepcopy(self.schema, memo)
        clone.__dict__ = deepcopy(self.__dict__, memo)
        return clone

    def __str__(self):
        return self.type


class PermissionMixIn(object):
    """mixin class for permissions handling"""
    def action_permissions(self, action):
        return self.permissions[action]

    def set_action_permissions(self, action, permissions):
        assert type(permissions) is tuple, (
            'permissions is expected to be a tuple not %s' % type(permissions))
        assert action in self.ACTIONS, ('%s not in %s' % (action, self.ACTIONS))
        self.permissions[action] = permissions

    def check_permission_definitions(self):
        """check permissions are correctly defined"""
        check_permission_definitions(self)


# Schema objects definition ###################################################


class EntitySchema(PermissionMixIn, ERSchema):
    """An entity has a type, a set of subject and or object relations
    the entity schema defines the possible relations for a given type and some
    constraints on those relations.
    """
    __implements__ = IEntitySchema

    ACTIONS = ('read', 'add', 'update', 'delete')
    field_checkers = BASE_CHECKERS
    field_converters = BASE_CONVERTERS

    # XXX set default values for those attributes on the class level since
    # they may be missing from schemas obtained by pyro
    _specialized_type = None
    _specialized_by = []

    def __init__(self, schema=None, rdef=None, *args, **kwargs):
        super(EntitySchema, self).__init__(schema, rdef, *args, **kwargs)
        if rdef is not None:
            # quick access to bounded relation schemas
            self.subjrels = {}
            self.objrels = {}
            self._specialized_type = rdef.specialized_type
            self._specialized_by = rdef.specialized_by
            self.final = self.type in BASE_TYPES
            self.permissions = rdef.__permissions__.copy()
            self._unique_together = getattr(rdef, '__unique_together__', [])
        else: # this happens during deep copy (cf. ERSchema.__deepcopy__)
            self._specialized_type = None
            self._specialized_by = []

    def check_unique_together(self):
        errors = []
        for unique_together in self._unique_together:
            for name in unique_together:
                try:
                    rschema = self.rdef(name, takefirst=True)
                except KeyError:
                    errors.append('no such attribute or relation %s' % name)
                else:
                    if not (rschema.final or rschema.rtype.inlined):
                        errors.append('%s is not an attribute or an inlined '
                                      'relation' % name)
        if errors:
            msg = 'invalid __unique_together__ specification for %s: %s' % (self, ', '.join(errors))
            raise BadSchemaDefinition(msg)

    def __repr__(self):
        return '<%s %s - %s>' % (self.type,
                                 [rs.type for rs in self.subject_relations()],
                                 [rs.type for rs in self.object_relations()])

    def _rehash(self):
        self.subjrels = rehash(self.subjrels)
        self.objrels = rehash(self.objrels)

    def advertise_new_add_permission(self):
        pass


    # schema building methods #################################################

    def add_subject_relation(self, rschema):
        """register the relation schema as possible subject relation"""
        self.subjrels[rschema] = rschema
        clear_cache(self, 'ordered_relations')
        clear_cache(self, 'meta_attributes')

    def add_object_relation(self, rschema):
        """register the relation schema as possible object relation"""
        self.objrels[rschema] = rschema

    def del_subject_relation(self, rtype):
        try:
            del self.subjrels[rtype]
            clear_cache(self, 'ordered_relations')
            clear_cache(self, 'meta_attributes')
        except KeyError:
            pass

    def del_object_relation(self, rtype):
        try:
            del self.objrels[rtype]
        except KeyError:
            pass

    # IEntitySchema interface #################################################

    ## navigation ######################

    def specializes(self):
        if self._specialized_type:
            return self.schema.eschema(self._specialized_type)
        return None

    def ancestors(self):
        specializes = self.specializes()
        ancestors = []
        while specializes:
            ancestors.append(specializes)
            specializes = specializes.specializes()
        return ancestors

    def specialized_by(self, recursive=True):
        eschema = self.schema.eschema
        subschemas = [eschema(etype) for etype in self._specialized_by]
        if recursive:
            for subschema in subschemas[:]:
                subschemas.extend(subschema.specialized_by(recursive=True))
        return subschemas

    def has_relation(self, rtype, role):
        if role == 'subject':
            return rtype in self.subjrels
        return rtype in self.objrels

    def subject_relations(self):
        """return a list of relations that may have this type of entity as
        subject
        """
        return self.subjrels.values()

    def object_relations(self):
        """return a list of relations that may have this type of entity as
        object
        """
        return self.objrels.values()

    def rdef(self, rtype, role='subject', targettype=None, takefirst=False):
        """return a relation definition schema for a relation of this entity type

        Notice that when targettype is not specified and the relation may lead
        to different entity types (ambiguous relation), one of them is picked
        randomly. If also takefirst is False, a warning will be emitted.
        """
        rschema = self.schema.rschema(rtype)
        if targettype is None:
            if role == 'subject':
                types = rschema.objects(self)
            else:
                types = rschema.subjects(self)
            if len(types) != 1 and not takefirst:
                warnings.warn('[yams 0.38] no targettype specified and there are several '
                              'relation definitions for rtype %s: %s. Yet you get the first '
                              'rdef.' % (rtype, [eschema.type for eschema in types]),
                              Warning, stacklevel=2)
            targettype = types[0]
        return rschema.role_rdef(self, targettype, role)

    @cached
    def ordered_relations(self):
        """return subject relations in an ordered way"""
        return sorted(self.subjrels.values(),
                      key=lambda x: x.rdef(self, x.objects(self)[0]).order)

    def relation_definitions(self, includefinal=False):
        """return an iterator on relation definitions

        if includefinal is false, only non attribute relation are returned

        a relation definition is a 3-uple :
        * schema of the (non final) relation
        * schemas of the possible destination entity types
        * a string telling if this is a 'subject' or 'object' relation
        """
        for rschema in self.ordered_relations():
            if includefinal or not rschema.final:
                yield rschema, rschema.objects(self), 'subject'
        for rschema in self.object_relations():
            yield rschema, rschema.subjects(self), 'object'

    def destination(self, rtype):
        """return the type or schema of entities related by the given subject relation

        `rtype` is expected to be a non ambiguous relation
        """
        rschema = self.subjrels[rtype]
        objtypes = rschema.objects(self.type)
        assert len(objtypes) == 1, (self.type, str(rtype),
                                    [str(ot) for ot in objtypes])
        # XXX return a RelationDefinitionSchema
        return objtypes[0]

    # attributes description ###########

    def attribute_definitions(self):
        """return an iterator on attribute definitions

        attribute relations are a subset of subject relations where the
        object's type is a final entity

        an attribute definition is a 2-uple :
        * schema of the (final) relation
        * schema of the destination entity type
        """
        for rschema in self.ordered_relations():
            if not rschema.final:
                continue
            # XXX return a RelationDefinitionSchema
            yield rschema, rschema.objects(self)[0]

    def main_attribute(self):
        """convenience method that returns the *main* (i.e. the first non meta)
        attribute defined in the entity schema
        """
        for rschema, _ in self.attribute_definitions():
            if not self.is_metadata(rschema):
                # XXX return a RelationDefinitionSchema
                return rschema

    def defaults(self):
        """return an iterator on (attribute name, default value)"""
        for rschema in self.subject_relations():
            if rschema.final:
                value = self.default(rschema)
                if value is not None:
                    # XXX return a RelationDefinitionSchema
                    yield rschema, value

    def default(self, rtype):
        """return the default value of a subject relation"""
        rdef = self.rdef(rtype, takefirst=True)
        default =  rdef.default
        if callable(default):
            default = default()
        if default is MARKER:
            default = None
        elif default is not None:
            return convert_default_value(rdef, default)
        return default

    def has_unique_values(self, rtype):
        """convenience method to check presence of the UniqueConstraint on a
        relation
        """
        return bool(self.rdef(rtype).constraint_by_class(UniqueConstraint))

    ## metadata attributes #############

    @cached
    def meta_attributes(self):
        """return a dictionnary defining meta-attributes:
        * key is an attribute schema
        * value is a 2-uple (metadata name, described attribute name)

        a metadata attribute is expected to be named using the following scheme:

          <described attribute name>_<metadata name>

        for instance content_format is the format metadata of the content
        attribute (if it exists).
        """
        metaattrs = {}
        for rschema, _ in self.attribute_definitions():
            try:
                attr, meta = rschema.type.rsplit('_', -1)
            except ValueError:
                continue
            if meta in KNOWN_METAATTRIBUTES and attr in self.subjrels:
                metaattrs[rschema] = (meta, attr)
        return metaattrs

    def has_metadata(self, attr, metadata):
        """return metadata's relation schema if this entity has the given
        `metadata` field for the given `attr` attribute
        """
        return self.subjrels.get('%s_%s' % (attr, metadata))

    def is_metadata(self, attr):
        """return a metadata for an attribute (None if unspecified)"""
        try:
            attr, metadata = str(attr).rsplit('_', 1)
        except ValueError:
            return None
        if metadata in KNOWN_METAATTRIBUTES and attr in self.subjrels:
            return (attr, metadata)
        return None

    # full text indexation control #####

    def indexable_attributes(self):
        """return the relation schema of attribtues to index"""
        for rschema in self.subject_relations():
            if rschema.final:
                try:
                    if self.rdef(rschema).fulltextindexed:
                        yield rschema
                except AttributeError:
                    # fulltextindexed is only available on String / Bytes
                    continue

    def fulltext_relations(self):
        """return the (name, role) of relations to index"""
        for rschema in self.subject_relations():
            if not rschema.final and rschema.fulltext_container == 'subject':
                yield rschema, 'subject'
        for rschema in self.object_relations():
            if rschema.fulltext_container == 'object':
                yield rschema, 'object'

    def fulltext_containers(self):
        """return relations whose extremity points to an entity that should
        contains the full text index content of entities of this type
        """
        for rschema in self.subject_relations():
            if rschema.fulltext_container == 'object':
                yield rschema, 'object'
        for rschema in self.object_relations():
            if rschema.fulltext_container == 'subject':
                yield rschema, 'subject'

    ## resource accessors ##############

    def is_subobject(self, strict=False, skiprels=()):
        """return True if this entity type is contained by another. If strict,
        return True if this entity type *must* be contained by another.
        """
        for rschema in self.object_relations():
            if (rschema, 'object') in skiprels:
                continue
            rdef = self.rdef(rschema, 'object', takefirst=True)
            if rdef.composite == 'subject':
                if not strict or rdef.cardinality[1] in '1+':
                    return True
        for rschema in self.subject_relations():
            if (rschema, 'subject') in skiprels:
                continue
            if rschema.final:
                continue
            rdef = self.rdef(rschema, 'subject', takefirst=True)
            if rdef.composite == 'object':
                if not strict or rdef.cardinality[0] in '1+':
                    return True
        return False

    ## validation ######################

    def check(self, entity, creation=False, _=None, relations=None):
        """check the entity and raises an ValidationError exception if it
        contains some invalid fields (ie some constraints failed)
        """
        if _ is not None:
            warnings.warn('[yams 0.36] _ argument is deprecated, remove it',
                          DeprecationWarning, stacklevel=2)
        _ = unicode
        errors = {}
        msgargs = {}
        i18nvalues = []
        relations = relations or self.subject_relations()
        for rschema in relations:
            if not rschema.final:
                continue
            aschema = self.destination(rschema)
            qname = role_name(rschema, 'subject')
            rdef = rschema.rdef(self, aschema)
            # don't care about rhs cardinality, always '*' (if it make senses)
            card = rdef.cardinality[0]
            assert card in '?1'
            required = card == '1'
            # check value according to their type
            try:
                value = entity[rschema]
            except KeyError:
                if creation and required:
                    # missing required attribute with no default on creation
                    # is not autorized
                    errors[qname] = _('required attribute')
                # on edition, missing attribute is considered as no changes
                continue
            # skip other constraint if value is None and None is allowed
            if value is None:
                if required:
                    errors[qname] = _('required attribute')
                continue
            rtype = rschema.type
            if not aschema.check_value(value):
                errors[qname] = _('incorrect value (%(KEY-value)r) for type "%(KEY-type)s"')
                msgargs[qname+'-value'] = value
                msgargs[qname+'-type'] = aschema.type
                i18nvalues.append(qname+'-type')
                if isinstance(value, str) and aschema == 'String':
                    errors[qname] += '; you might want to try unicode'
                continue
            # ensure value has the correct python type
            nvalue = aschema.convert_value(value)
            if nvalue != value:
                # don't change what's has not changed, who knows what's behind
                # this <entity> thing
                entity[rschema] = value = nvalue
            # check arbitrary constraints
            for constraint in rdef.constraints:
                if not constraint.check(entity, rschema, value):
                    msg, args = constraint.failed_message(qname, value)
                    errors[qname] = msg
                    msgargs.update(args)
                    break
        if errors:
            raise ValidationError(entity, errors, msgargs, i18nvalues)

    def check_value(self, value):
        """check the value of a final entity (ie a const value)"""
        return self.field_checkers[self](self, value)

    def convert_value(self, value):
        """check the value of a final entity (ie a const value)"""
        try:
            return self.field_converters[self](value)
        except KeyError:
            return value

    def vocabulary(self, rtype):
        """backward compat return the vocabulary of a subject relation
        """
        cstr = self.rdef(rtype).constraint_by_interface(IVocabularyConstraint)
        if cstr is None:
            raise AssertionError('field %s of entity %s has no vocabulary' %
                                 (rtype, self))
        return cstr.vocabulary()



class RelationSchema(ERSchema):
    """A relation is a named and oriented link between two entities.
    A relation schema defines the possible types of both extremities.

    Cardinality between the two given entity's type is defined
    as a 2 characters string where each character is one of:
     - 1 <-> 1..1 <-> one and only one
     - ? <-> 0..1 <-> zero or one
     - + <-> 1..n <-> one or more
     - * <-> 0..n <-> zero or more
    """

    __implements__ = IRelationSchema

    def __init__(self, schema=None, rdef=None, **kwargs):
        if rdef is not None:
            # if this relation is symmetric/inlined
            self.symmetric = rdef.symmetric or False
            self.inlined = rdef.inlined or False
            # if full text content of subject/object entity should be added
            # to other side entity (the container)
            self.fulltext_container = rdef.fulltext_container or None
            # if this relation is an attribute relation
            self.final = False
            # mapping to subject/object with schema as key
            self._subj_schemas = {}
            self._obj_schemas = {}
            # relation properties
            self.rdefs = {}
        super(RelationSchema, self).__init__(schema, rdef, **kwargs)

    def __repr__(self):
        return '<%s [%s]>' % (self.type,
                              '; '.join('%s,%s' % (s.type, o.type)
                                        for (s, o), props in self.rdefs.items()))

    def _rehash(self):
        self._subj_schemas = rehash(self._subj_schemas)
        self._obj_schemas = rehash(self._obj_schemas)
        self.rdefs = rehash(self.rdefs)

    # schema building methods #################################################

    def update(self, subjschema, objschema, rdef):
        """Allow this relation between the two given types schema"""
        if subjschema.final:
            msg = 'type %s can\'t be used as subject in a relation' % subjschema
            raise BadSchemaDefinition(msg)
        # check final consistency:
        # * a final relation only points to final entity types
        # * a non final relation only points to non final entity types
        final = objschema.final
        for eschema in self.objects():
            if eschema is objschema:
                continue
            if final != eschema.final:
                if final:
                    frschema, nfrschema = objschema, eschema
                else:
                    frschema, nfrschema = eschema, objschema
                msg = "ambiguous relation %s: %s is final but not %s" % (
                    self.type, frschema, nfrschema)
                raise BadSchemaDefinition(msg)
        constraints = getattr(rdef, 'constraints', None)
        if constraints:
            for cstr in constraints:
                cstr.check_consistency(subjschema, objschema, rdef)
        if (subjschema, objschema) in self.rdefs and self.symmetric:
            return
        # update our internal struct
        if final:
            assert not self.symmetric, 'no sense on final relation'
            assert not self.inlined, 'no sense on final relation'
            assert not self.fulltext_container, 'no sense on final relation'
        self.final = final
        rdefs = self.init_rproperties(subjschema, objschema, rdef)
        self._add_rdef(rdefs)
        return rdefs

    def _add_rdef(self, rdef):
        # update our internal struct
        self.rdefs[(rdef.subject, rdef.object)] = rdef
        self._update(rdef.subject, rdef.object)
        if self.symmetric:
            self._update(rdef.object, rdef.subject)
            if rdef.object != rdef.subject:
                self.rdefs[(rdef.object, rdef.subject)] = rdef
        if self.inlined and rdef.cardinality[0] in '*+':
            raise BadSchemaDefinition(
                'inlined relation %s can\'t have multiple cardinality for its '
                'subject' % rdef)
        # update entity types schema
        rdef.subject.add_subject_relation(self)
        if self.symmetric:
            rdef.object.add_subject_relation(self)
        else:
            rdef.object.add_object_relation(self)

    def _update(self, subjectschema, objectschema):
        objtypes = self._subj_schemas.setdefault(subjectschema, [])
        if not objectschema in objtypes:
            objtypes.append(objectschema)
        subjtypes = self._obj_schemas.setdefault(objectschema, [])
        if not subjectschema in subjtypes:
            subjtypes.append(subjectschema)

    def del_relation_def(self, subjschema, objschema, _recursing=False):
        try:
            self._subj_schemas[subjschema].remove(objschema)
            if len(self._subj_schemas[subjschema]) == 0:
                del self._subj_schemas[subjschema]
                subjschema.del_subject_relation(self)
        except (ValueError, KeyError):
            pass
        try:
            self._obj_schemas[objschema].remove(subjschema)
            if len(self._obj_schemas[objschema]) == 0:
                del self._obj_schemas[objschema]
                objschema.del_object_relation(self)
        except (ValueError, KeyError):
            pass
        try:
            del self.rdefs[(subjschema, objschema)]
        except KeyError:
            pass
        try:
            if self.symmetric and subjschema != objschema and not _recursing:
                self.del_relation_def(objschema, subjschema, True)
        except KeyError:
            pass
        if not self._obj_schemas or not self._subj_schemas:
            assert not self._obj_schemas and not self._subj_schemas
            return True
        return False

    # relation definitions properties handling ################################

    # XXX move to RelationDefinitionSchema

    def init_rproperties(self, subject, object, buildrdef):
        key = subject, object
        # raise an error if already defined unless the defined relalation has
        # been infered, in which case we may want to replace it
        if key in self.rdefs and not self.rdefs[key].infered:
            msg = '(%s, %s) already defined for %s' % (subject, object, self)
            raise BadSchemaDefinition(msg)
        self.rdefs[key] = rdef = RelationDefinitionSchema(subject, self, object,
                                                          buildrdef.package)
        for prop, default in rdef.rproperties().iteritems():
            rdefval = getattr(buildrdef, prop, MARKER)
            if rdefval is MARKER:
                if prop == 'permissions':
                    rdefval = default = buildrdef.get_permissions(self.final).copy()
                if prop == 'cardinality':
                    default = (object in BASE_TYPES) and '?1' or '**'
            else:
                default = rdefval
            setattr(rdef, prop, default)
        return rdef

    # IRelationSchema interface ###############################################

    def associations(self):
        """return a list of (subject, [objects]) defining between which types
        this relation may exists
        """
        # XXX deprecates in favor of iter_rdefs() ?
        return self._subj_schemas.items()

    def subjects(self, etype=None):
        """Return a list of entity schemas which can be subject of this relation.

        If etype is not None, return a list of schemas which can be subject of
        this relation with etype as object.

        :raise `KeyError`: if etype is not a subject entity type.
        """
        if etype is None:
            return tuple(self._subj_schemas)
        try:
            return tuple(self._obj_schemas[etype])
        except KeyError:
            raise KeyError("%s does not have %s as object" % (self, etype))

    def objects(self, etype=None):
        """Return a list of entity schema which can be object of this relation.

        If etype is not None, return a list of schemas which can be object of
        this relation with etype as subject.

        :raise `KeyError`: if etype is not an object entity type.
        """
        if etype is None:
            return tuple(self._obj_schemas)
        try:
            return tuple(self._subj_schemas[etype])
        except KeyError:
            raise KeyError("%s does not have %s as subject" % (self, etype))

    def targets(self, etype=None, role='subject'):
        """return possible target types with <etype> as <x>"""
        if role == 'subject':
            return self.objects(etype)
        return self.subjects(etype)

    def rdef(self, subject, object):
        """return the properties dictionary of a relation"""
        try:
            return self.rdefs[(subject, object)]
        except KeyError:
            raise KeyError('%s %s %s' % (subject, self, object))

    def role_rdef(self, etype, ttype, role):
        if role == 'subject':
            return self.rdefs[(etype, ttype)]
        return self.rdefs[(ttype, etype)]

    def check_permission_definitions(self):
        """check permissions are correctly defined"""
        for rdef in self.rdefs.itervalues():
            rdef.check_permission_definitions()



class RelationDefinitionSchema(PermissionMixIn):
    """a relation definition is fully caracterized relation, eg

         <subject type> <relation type> <object type>
    """

    _RPROPERTIES = {'cardinality': None,
                    'constraints': (),
                    'order': 9999,
                    'description': '',
                    'infered': False,
                    'permissions': None}
    _NONFINAL_RPROPERTIES = {'composite': None}
    _FINAL_RPROPERTIES = {'default': None,
                          'uid': False,
                          'indexed': False}
    # Use a TYPE_PROPERTIES dictionnary to store type-dependant parameters.
    BASE_TYPE_PROPERTIES = {'String': {'fulltextindexed': False,
                                       'internationalizable': False},
                            'Bytes': {'fulltextindexed': False}}

    @classmethod
    def ALL_PROPERTIES(cls):
        return set(chain(cls._RPROPERTIES,
                         cls._NONFINAL_RPROPERTIES,
                         cls._FINAL_RPROPERTIES,
                         *cls.BASE_TYPE_PROPERTIES.itervalues()))

    def __init__(self, subject, rtype, object, package, values=None):
        if values is not None:
            self.update(values)
        self.subject = subject
        self.rtype = rtype
        self.object = object
        self.package = package

    @property
    def ACTIONS(self):
        if self.rtype.final:
            return ('read', 'add', 'update')
        else:
            return ('read', 'add', 'delete')

    def update(self, values):
        # XXX check we're copying existent properties
        self.__dict__.update(values)

    def __str__(self):
        if self.object.final:
            return 'attribute %s.%s[%s]' % (
            self.subject, self.rtype, self.object)
        return 'relation %s %s %s' % (
            self.subject, self.rtype, self.object)

    def __repr__(self):
        return '<%s at @%#x>' % (self, id(self))

    def as_triple(self):
        return (self.subject, self.rtype, self.object)

    def advertise_new_add_permission(self):
        """handle backward compatibility with pre-add permissions

        * if the update permission was () [empty tuple], use the
          default attribute permissions for `add`

        * else copy the `update` rule for `add`
        """
        if not 'add' in self.permissions:
            from yams.buildobjs import DEFAULT_ATTRPERMS
            if self.permissions['update'] == ():
                defaultaddperms = DEFAULT_ATTRPERMS['add']
            else:
                defaultaddperms = self.permissions['update']
            self.permissions['add'] = defaultaddperms
            warnings.warn('[yams 0.39] %s: new "add" permissions on attribute '
                          'set to %s by default, but you must make it explicit' %
                          (self, defaultaddperms), DeprecationWarning)


    @classmethod
    def rproperty_defs(cls, desttype):
        """return a dictionary mapping property name to its definition for each
        allowable properties when the relation has `desttype` as target entity's
        type
        """
        propdefs = cls._RPROPERTIES.copy()
        if desttype not in BASE_TYPES:
            propdefs.update(cls._NONFINAL_RPROPERTIES)
        else:
            propdefs.update(cls._FINAL_RPROPERTIES)
            propdefs.update(cls.BASE_TYPE_PROPERTIES.get(desttype, {}))
        return propdefs

    def rproperties(self):
        """same as .rproperty_defs class method, but for instances (hence
        destination is known to be self.object).
        """
        return self.rproperty_defs(self.object)

    def get(self, key, default=None):
        return getattr(self, key, default)

    @property
    def final(self):
        return self.rtype.final

    def dump(self, subject, object):
        return RelationDefinitionSchema(subject, self.rtype, object,
                                        self.package,
                                        self.__dict__)

    def role_cardinality(self, role):
        return self.cardinality[role == 'object']

    def constraint_by_type(self, cstrtype):
        for cstr in self.constraints:
            if cstr.type() == cstrtype:
                return cstr
        return None

    def constraint_by_class(self, cls):
        for cstr in self.constraints:
            if isinstance(cstr, cls):
                return cstr
        return None

    def constraint_by_interface(self, iface):
        for cstr in self.constraints:
            if implements(cstr, iface):
                return cstr
        return None


class Schema(object):
    """set of entities and relations schema defining the possible data sets
    used in an application


    :type name: str
    :ivar name: name of the schema, usually the application identifier

    :type base: str
    :ivar base: path of the directory where the schema is defined
    """
    __implements__ = ISchema
    entity_class = EntitySchema
    relation_class = RelationSchema
    # relation that should not be infered according to entity type inheritance
    no_specialization_inference = ()

    def __init__(self, name, construction_mode='strict'):
        super(Schema, self).__init__()
        self.name = name
        # with construction_mode != 'strict', no error when trying to add a
        # relation using an undefined entity type, simply log the error
        self.construction_mode = construction_mode
        self._entities = {}
        self._relations = {}

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._rehash()

    def _rehash(self):
        """rehash schema's internal structures"""
        for eschema in self._entities.values():
            eschema._rehash()
        for rschema in self._relations.values():
            rschema._rehash()

    def get(self, name, default=None):
        try:
            return self[name]
        except KeyError:
            return default

    def __getitem__(self, name):
        try:
            return self.eschema(name)
        except KeyError:
            return self.rschema(name)

    def __contains__(self, name):
        try:
            self[name]
        except KeyError:
            return False
        return True

    # schema building methods #################################################

    def add_entity_type(self, edef):
        """Add an entity schema definition for an entity's type.

        :type edef: str
        :param edef: the name of the entity type to define

        :raise `BadSchemaDefinition`: if the entity type is already defined
        :rtype: `EntitySchema`
        :return: the newly created entity schema instance
        """
        etype = edef.name
        if etype in self._entities:
            msg = "entity type %s is already defined" % etype
            raise BadSchemaDefinition(msg)
        eschema = self.entity_class(self, edef)
        self._entities[etype] = eschema
        return eschema

    def rename_entity_type(self, oldname, newname):
        """renames an entity type and update internal structures accordingly
        """
        eschema = self._entities.pop(oldname)
        eschema.type = newname
        self._entities[newname] = eschema
        # rebuild internal structures since eschema's hash value has changed
        self._rehash()


    def add_relation_type(self, rtypedef):
        rtype = rtypedef.name
        if rtype in self._relations:
            msg = "relation type %s is already defined" % rtype
            raise BadSchemaDefinition(msg)
        rschema = self.relation_class(self, rtypedef)
        self._relations[rtype] = rschema
        return rschema

    def add_relation_def(self, rdef):
        """build a part of a relation schema:
        add a relation between two specific entity's types

        :rtype: RelationSchema
        :return: the newly created or simply completed relation schema
        """
        rtype = rdef.name
        try:
            rschema = self.rschema(rtype)
        except KeyError:
            return self._building_error('using unknown relation type in %s',
                                        rdef)
        try:
            subjectschema = self.eschema(rdef.subject)
        except KeyError:
            return self._building_error('using unknown type %r in relation %s',
                                        rdef.subject, rtype)
        try:
            objectschema = self.eschema(rdef.object)
        except KeyError:
            return self._building_error("using unknown type %r in relation %s",
                                        rdef.object, rtype)
        return rschema.update(subjectschema, objectschema, rdef)

    def _building_error(self, msg, *args):
        if self.construction_mode == 'strict':
            raise BadSchemaDefinition(msg % args)
        self.critical(msg, *args)

    def del_relation_def(self, subjtype, rtype, objtype):
        subjschema = self.eschema(subjtype)
        objschema = self.eschema(objtype)
        rschema = self.rschema(rtype)
        if rschema.del_relation_def(subjschema, objschema):
            del self._relations[rtype]

    def del_relation_type(self, rtype):
        # XXX don't iter directly on the dictionary since it may be changed
        # by del_relation_def
        for subjtype, objtype in self.rschema(rtype).rdefs.keys():
            self.del_relation_def(subjtype, rtype, objtype)
        if not self.rschema(rtype).rdefs:
            del self._relations[rtype]

    def del_entity_type(self, etype):
        eschema = self._entities[etype]
        for rschema in eschema.subjrels.values():
            for objtype in rschema.objects(etype):
                self.del_relation_def(eschema, rschema, objtype)
        for rschema in eschema.objrels.values():
            for subjtype in rschema.subjects(etype):
                self.del_relation_def(subjtype, rschema, eschema)
        if eschema.specializes():
            eschema.specializes()._specialized_by.remove(eschema)
        if eschema.specialized_by():
            raise Exception("can't remove entity type %s used as parent class by %s" %
                            (eschema, ','.join(str(et) for et in eschema.specialized_by())))
        del self._entities[etype]

    def infer_specialization_rules(self):
        for rschema in self.relations():
            if rschema in self.no_specialization_inference:
                continue
            for (subject, object), rdef in rschema.rdefs.items():
                subjeschemas = [subject] + subject.specialized_by(recursive=True)
                objeschemas = [object] + object.specialized_by(recursive=True)
                for subjschema in subjeschemas:
                    for objschema in objeschemas:
                        # don't try to add an already defined relation
                        if (subjschema, objschema) in rschema.rdefs:
                            continue
                        thisrdef = rdef.dump(subjschema, objschema)
                        thisrdef.infered = True
                        rschema._add_rdef(thisrdef)

    def remove_infered_definitions(self):
        """remove any infered definitions added by
        `infer_specialization_rules`
        """
        for rschema in self.relations():
            if rschema.final:
                continue
            for (subject, object), rdef in rschema.rdefs.items():
                if rdef.infered:
                    self.del_relation_def(subject, rschema, object)

    def rebuild_infered_relations(self):
        """remove any infered definitions and rebuild them"""
        self.remove_infered_definitions()
        self.infer_specialization_rules()

    # ISchema interface #######################################################

    def entities(self):
        """return a list of possible entity's type

        :rtype: list
        :return: defined entity's types (str) or schemas (`EntitySchema`)
        """
        return self._entities.values()

    def has_entity(self, etype):
        """return true the type is defined in the schema

        :type etype: str
        :param etype: the entity's type

        :rtype: bool
        :return:
          a boolean indicating whether this type is defined in this schema
        """
        return etype in self._entities

    def eschema(self, etype):
        """return the entity's schema for the given type

        :rtype: `EntitySchema`
        :raise `KeyError`: if the type is not defined as an entity
        """
        try:
            return self._entities[etype]
        except KeyError:
            if isinstance(etype, tuple):
                etype = list(etype)
            raise KeyError('No entity named %s in schema' % etype)

    def relations(self):
        """return the list of possible relation'types

        :rtype: list
        :return: defined relation's types (str) or schemas (`RelationSchema`)
        """
        return self._relations.values()

    def has_relation(self, rtype):
        """return true the relation is defined in the schema

        :type rtype: str
        :param rtype: the relation's type

        :rtype: bool
        :return:
          a boolean indicating whether this type is defined in this schema
        """
        return rtype in self._relations

    def rschema(self, rtype):
        """return the relation schema for the given type

        :rtype: `RelationSchema`
        """
        try:
            return self._relations[rtype]
        except KeyError:
            raise KeyError('No relation named %s in schema'%rtype)


import logging
from logilab.common.logging_ext import set_log_methods
LOGGER = logging.getLogger('yams')
set_log_methods(Schema, LOGGER)



# copyright 2004-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
from logilab.common.testlib import TestCase, unittest_main
from yams.xy import XYRegistry

class XYTC(TestCase):
    def test(self):
        xy = XYRegistry()
        xy.register_prefix('dc', 'http://purl.org/dc/elements/1.1/')
        xy.register_prefix('foaf', 'http://xmlns.com/foaf/0.1/')
        xy.register_prefix('doap', 'http://usefulinc.com/ns/doap#')

        xy.add_equivalence('creation_date', 'dc:date')
        xy.add_equivalence('created_by', 'dc:creator')
        xy.add_equivalence('description', 'dc:description')
        xy.add_equivalence('CWUser', 'foaf:Person')
        xy.add_equivalence('CWUser login', 'dc:title')
        xy.add_equivalence('CWUser surname', 'foaf:Person foaf:name')
        xy.add_equivalence('Project', 'doap:Project')
        xy.add_equivalence('Project name', 'dc:title')
        xy.add_equivalence('Project name', 'doap:Project doap:name')
        xy.add_equivalence('Project creation_date', 'doap:Project doap:created')
        xy.add_equivalence('Project created_by CWUser', 'doap:Project doap:maintainer foaf:Person')
        xy.add_equivalence('Version', 'doap:Version')
        xy.add_equivalence('Version name', 'doap:Version doap:name')
        xy.add_equivalence('Version creation_date', 'doap:Version doap:created')
        xy.add_equivalence('Version num', 'doap:Version doap:revision')

        self.assertEqual(xy.yeq('doap:Project', isentity=True), ['Project'])
        self.assertEqual(xy.yeq('dc:title'), [('CWUser', 'login', '*'),
                                               ('Project', 'name', '*')])
        self.assertEqual(xy.yeq('doap:Project doap:maintainer'),
                          [('Project', 'created_by', 'CWUser')])

if __name__ == '__main__':
    unittest_main()



# copyright 2004-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""XML vocabularies <-> Yams schema mapping"""

from warnings import warn

class UnsupportedVocabulary(Exception): pass

class XYRegistry(object):
    """registry of equivalence to translate yams into various xml schema such
    as dublin core, foaf, doap...

    """
    def __init__(self):
        self.prefixes = {}
        self._y2x = {}
        self._x2y = {}

    def register_prefix(self, prefix, xmlns, overwrite=False):
        if ':' in prefix:
            warn('[yams 0.30] register_prefix arguments has been switched to '
                 '(prefix, xmlns)', DeprecationWarning, stacklevel=2)
            prefix, xmlns = xmlns, prefix
        if not overwrite and prefix in self.prefixes \
                and self.prefixes.get(prefix) != xmlns:
            raise Exception('prefix %r already defined with different value'
                            % prefix)
        self.prefixes[prefix] = xmlns

    def _norm_yams_key(self, yamssnippet):
        parts = yamssnippet.split(' ')
        if len(parts) == 1:
            if parts[0][0].islower():
                return '*', parts[0], '*'
            return parts[0]
        if len(parts) == 2:
            parts.append('*')
        assert len(parts) == 3
        return tuple(parts)

    def _norm_xml_key(self, xmlsnippet, isentity=False):
        parts = []
        for xmlsnippet in xmlsnippet.split(' '):
            pfx, tag = xmlsnippet.rsplit(':', 1)
            xmlns = self.prefixes.get(pfx, pfx)
            parts.append( (xmlns, tag) )
        if isentity:
            assert len(parts) == 1
            return parts[0]
        if len(parts) == 1:
            return '*', parts[0], '*'
        if len(parts) == 2:
            parts.append('*')
        assert len(parts) == 3
        return tuple(parts)

    def add_equivalence(self, yamssnippet, xmlsnippet):
        ykey = self._norm_yams_key(yamssnippet)
        isentity = not isinstance(ykey, tuple)
        xkey = self._norm_xml_key(xmlsnippet, isentity=isentity)
        self._y2x.setdefault(ykey, []).append(xkey)
        self._x2y.setdefault(xkey, []).append(ykey)
        if not isentity:
            for dict, key, value in ((self._x2y, xkey, ykey),
                                     (self._y2x, ykey, xkey)):
                if key[0] != '*':
                    wkey = ('*',) + key[1:]
                    dict.setdefault(wkey, []).append(value)
                if key[2] != '*':
                    wkey = key[:2] + ('*',)
                    dict.setdefault(wkey, []).append(value)

    def remove_equivalence(self, yamssnippet, xmlsnippet):
        ykey = self._norm_yams_key(yamssnippet)
        isentity = not isinstance(ykey, tuple)
        xkey = self._norm_xml_key(xmlsnippet, isentity=isentity)
        self._y2x.setdefault(ykey, []).remove(xkey)
        self._x2y.setdefault(xkey, []).remove(ykey)
        if not isentity:
            for dict, key, value in ((self._x2y, xkey, ykey),
                                     (self._y2x, ykey, xkey)):
                if key[0] != '*':
                    wkey = ('*',) + key[1:]
                    dict.setdefault(wkey, []).remove(value)
                if key[2] != '*':
                    wkey = key[:2] + ('*',)
                    dict.setdefault(wkey, []).remove(value)

    def yeq(self, xmlsnippet, isentity=False):
        key = self._norm_xml_key(xmlsnippet, isentity)
        try:
            return self._x2y[key]
        except KeyError:
            raise UnsupportedVocabulary(key)

    def xeq(self, yamssnippet):
        key = self._norm_yams_key(yamssnippet)
        try:
            return self._y2x[key]
        except KeyError:
            raise UnsupportedVocabulary(key)

XY = XYRegistry()
register_prefix = XY.register_prefix
add_equivalence = XY.add_equivalence
remove_equivalence = XY.remove_equivalence
yeq = XY.yeq
xeq = XY.xeq



# copyright 2004-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""YAMS exception classes"""

__docformat__ = "restructuredtext en"

class SchemaError(Exception):
    """base class for schema exceptions"""

    def __str__(self):
        return unicode(self).encode('utf8')


class UnknownType(SchemaError):
    """using an unknown entity type"""

    msg = 'Unknown type %s'

    def __unicode__(self):
        return self.msg % self.args


class BadSchemaDefinition(SchemaError):
    """error in the schema definition

    instance attributes:
    * filename is the source file where the exception was raised
    * lineno is the line number where the exception was raised
    * line is the actual line in text form
    """

    msg = '%s line %s: %s'

    def __get_filename(self):
        if len(self.args) > 1:
            return self.args[0]
        else:
            return None
    filename = property(__get_filename)

    def __unicode__(self):
        msgs = []
        args_offset = 0
        if self.filename is not None:
            msgs.append(self.filename)
            args_offset += 1
            msgs.append(': ')
        msgs.append(' '.join(self.args[args_offset:]))
        return ''.join(msgs)


class ValidationError(SchemaError):
    """Validation error details the reason(s) why the validation failed.

    Arguments are:

    * `entity`: the entity that could not be validated; actual type depends on
      the client library

    * `errors`: errors dictionary, None key used for global error, other keys
      should be attribute/relation of the entity, qualified as subject/object
      using :func:`yams.role_name`.  Values are the message associated to the
      keys, and may include interpolation string starting with '%(KEY-' where
      'KEY' will be replaced by the associated key once the message has been
      translated. This allows predictable/translatable message and avoid args
      conflict if used for several keys.

    * `msgargs`: dictionary of substitutions to be inserted in error
      messages once translated (only if msgargs is given)

    * `i18nvalues`: list of keys in msgargs whose value should be translated

    Translation will be done **in-place** by calling :meth:`translate`.
    """

    def __init__(self, entity, errors, msgargs=None, i18nvalues=None):
        # set args so ValidationError are serializable through pyro
        SchemaError.__init__(self, entity, errors)
        self.entity = entity
        assert isinstance(errors, dict), 'validation errors must be a dict'
        self.errors = errors
        self.msgargs = msgargs
        self.i18nvalues = i18nvalues
        self._translated = False

    def __unicode__(self):
        if self._translated:
            errors = self.errors
        else:
            errors = dict(self._translated_errors(unicode))
        if len(errors) == 1:
            attr, error = errors.items()[0]
            return u'%s (%s): %s' % (self.entity, attr, error)
        errors = '\n'.join('* %s: %s' % (k, v) for k, v in errors.items())
        return u'%s:\n%s' % (self.entity, errors)

    def translate(self, _):
        """Translate and interpolate messsages in the errors dictionary, using
        the given translation function.

        If no substitution has been given, suppose msg is already translated for
        bw compat, so no translation occurs.

        This method may only be called once.
        """
        assert not self._translated
        self._translated = True
        if self.msgargs is not None:
            if self.i18nvalues:
                for key in self.i18nvalues:
                    self.msgargs[key] = _(self.msgargs[key])
            self.errors = dict(self._translated_errors(_))

    def _translated_errors(self, _):
        for key, msg in self.errors.items():
            msg = _(msg)
            if key is not None:
                msg = msg.replace('%(KEY-', '%('+key+'-')
            yield key, msg % self.msgargs

